// Copyright 2013 SNCF. All rights reserved.
package com.sncf.checker;

import java.util.Iterator;
import java.util.List;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.sncf.checker.bo.Arrival;
import com.sncf.checker.bo.Event;
import com.sncf.checker.bo.Fill;
import com.sncf.checker.bo.JointArrDep;
import com.sncf.checker.bo.JointArrDepEntry;
import com.sncf.checker.bo.Train;
import com.sncf.checker.common.ROADEFException;
import com.sncf.checker.io.ArrivalDAO;
import com.sncf.checker.io.DAOFactory;
import com.sncf.checker.io.TrainDAO;

/**
 * Simultaneous events of assembled trains.
 *
 * When trains are assembled, they must have the same events in their
 * associated schedules.
 *
 * @author YDS
 *
 */
public class AssembledTrainsSimultaneousEventConstraint implements Constraint {

    /** Interface to Data layer. */
    private final DAOFactory factory;

    /**
     * Constructor.
     *
     * @param factory dao factory to set
     */
    public AssembledTrainsSimultaneousEventConstraint(
            final DAOFactory factory) {
        this.factory = factory;
    }

    /** {@inheritDoc} */
    public final boolean validate() throws ROADEFException {
        ArrivalDAO arrivalDAO = factory.getArrivalDAO();
        TrainDAO trainDAO = factory.getTrainDAO();
        boolean valid = true;
        Logger logger = Logger.getLogger("logger");

        List<Train> trainList = trainDAO.loadAll();
        for (Train train : trainList) {
            Iterator<Fill> current = train.getFill().iterator();
            TreeMap<String, Iterator<Fill>> assembled =
                    new TreeMap<String, Iterator<Fill>>();
            Arrival arrival = train.getArrival();
            if (arrival != null) {
                JointArrDep joint = arrival.getJointArrDep();
                if (joint != null) {
                    for (JointArrDepEntry entry : joint.getEntryList()) {
                        Arrival jointArrival =
                                arrivalDAO.loadByName(entry.getArrDep());
                        if (jointArrival.getTrain() != train
                                && !jointArrival.getTrain().getFill().isEmpty()) {
                           assembled.put(jointArrival.getTrain().getName(),
                                   jointArrival.getTrain().getFill().iterator());
                        }
                    }
                }
            }
            while (current.hasNext()) {
                Fill fill = current.next();
                for (Iterator<Fill> itr : assembled.values()) {
                    if (itr.hasNext()) {
                        Fill associated = itr.next();
                        if (fill.getDate().getTime()
                                != associated.getDate().getTime()) {
                            logger.log(Level.INFO, "Simultaneous events of "
                                    + "assembled trains error : "
                                    + train.getName() + " and "
                                    + associated.getTrain().getName()
                                    + " have events with different times ("
                                    + fill.getDate().toString()
                                    + " and " + associated.getDate().toString()
                                    + " respectively).");
                            valid = false;
                            continue;
                        }
                        if (!fill.getEvent().equals(associated.getEvent())) {
                            logger.log(Level.INFO, "Simultaneous events of "
                                    + "assembled trains error : "
                                    + train.getName() + " and "
                                    + associated.getTrain().getName()
                                    + " have events with different types at "
                                    + fill.getDate().toString()
                                    + " ("  + fill.getEvent().toString() + " "
                                    + associated.getEvent().toString()
                                    + " respectively).");
                            valid = false;
                            continue;
                        }
                        if (fill.getResource() != associated.getResource()) {
                            logger.log(Level.INFO, "Simultaneous events of "
                                    + "assembled trains error : "
                                    + train.getName() + " and "
                                    + associated.getTrain().getName()
                                    + " have events with different resources "
                                    + "at " + fill.getDate().toString()
                                    + " ("  + fill.getResource().getName() + " "
                                    + associated.getResource().getName()
                                    + " respectively).");
                            valid = false;
                            continue;
                        }
                        if (!fill.getGate().equals(associated.getGate())) {
                            logger.log(Level.INFO, "Simultaneous events of "
                                    + "assembled trains error : "
                                    + train.getName() + " and "
                                    + associated.getTrain().getName()
                                    + " have events with different gates at "
                                    + fill.getDate().toString()
                                    + " ("  + fill.getGate() + " "
                                    + associated.getGate()
                                    + " respectively).");
                            valid = false;
                            continue;
                        }
                        if (!fill.getEvent().equals(Event.Arrival)
                                && !fill.getEvent().equals(Event.Departure)
                                && !fill.getEvent().equals(Event.EndDisjunction)) {
                            if (!fill.getComplement().equals(
                                    associated.getComplement())) {
                                logger.log(Level.INFO, "Simultaneous events of "
                                        + "assembled trains error : "
                                        + train.getName() + " and "
                                        + associated.getTrain().getName()
                                        + " have events with different "
                                        + "complements at "
                                        + fill.getDate().toString()
                                        + " ("  + fill.getComplement() + " "
                                        + associated.getComplement()
                                        + " respectively).");
                                valid = false;
                                continue;
                            }
                        }
                    } else {
                        logger.log(Level.INFO, "Simultaneous events of "
                                + "assembled trains error : "
                                + train.getName() + " has an associated "
                                + "train with different events at "
                                + fill.getDate().toString() + ".");
                        valid = false;
                        continue;
                    }
                }
                if (fill.getEvent().equals(Event.EndJunction)) {
                    String[] complement = fill.getComplement().split("\\+");
                    for (String s : complement) {
                        if (!assembled.containsKey(s)
                                && !s.equals(train.getName())) {
                            Train tr = trainDAO.loadByName(s);
                            if (tr != null) {
                                Iterator<Fill> itr =
                                        getBeginningAssembledPeriod(fill, tr);
                                if (itr != null) {
                                    assembled.put(tr.getName(), itr);
                                } else {
                                    logger.log(Level.INFO, "Simultaneous "
                                            + "events of assembled trains "
                                            + "error : " + train.getName()
                                            + " has an associated train ("
                                            + s + ") with different events at "
                                            + fill.getDate().toString() + ".");
                                    valid = false;
                                }
                            }
                        }
                    }
                } else if (fill.getEvent().equals(Event.EndDisjunction)) {
                    TreeMap<String, Iterator<Fill>> tmp =
                            new TreeMap<String, Iterator<Fill>>();
                    String[] complement = fill.getComplement().split("\\+");
                    for (String s : complement) {
                        if (!s.equals(train.getName())) {
                            Iterator<Fill> itr = assembled.get(s);
                            if (itr != null) {
                                tmp.put(s, itr);
                            } else {
                                logger.log(Level.INFO, "Simultaneous events of "
                                        + "assembled trains error : "
                                        + train.getName()
                                        + " has an unknown associated train ("
                                        + s + ") at " + fill.getDate().toString()
                                        + ".");
                                valid = false;
                            }
                        }
                    }
                    assembled = tmp;
                }
            }
        }
        return valid;
    }

    /**
     * Get the beginning Fill of the assembled train.
     *
     * @param fill searched fill
     * @param train assembled train
     * @return iterator on the beginning fill
     */
    private Iterator<Fill> getBeginningAssembledPeriod(final Fill fill
            , final Train train) throws ROADEFException {
        Iterator<Fill> current = train.getFill().iterator();
        while (current.hasNext()) {
            Fill associated = current.next();
            if (fill.getDate().getTime() == associated.getDate().getTime()) {
                if (fill.getEvent().equals(associated.getEvent())) {
                    if (fill.getComplement().equals(
                            associated.getComplement())) {
                        return current;
                    }
                }
            }
        }
        return null;
    }
}
