// Copyright 2013 SNCF. All rights reserved.
package com.sncf.checker;

import java.io.IOException;
import java.util.logging.FileHandler;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.logging.SimpleFormatter;

import com.sncf.checker.common.Datetime;
import com.sncf.checker.common.ROADEFException;
import com.sncf.checker.io.DAOFactory;
import com.sncf.checker.io.FlatFileDAOFactory;

public class Checker {

    /**
     * @param args
     */
    public static void main(String[] args) {
        String date = "12/06/2014";
        String version = "2.3";
        String welcome = "This is the ROADEF/EURO Challenge 2014 checker for the final phase, version " + version
                + " of " + date + ".";
        System.out.println(welcome);

        if (args.length != 2) {
            printUsage();
            System.exit(1);
        }

        float score;
        boolean feasible;


        Logger logger = Logger.getLogger("logger");
        try {

            try {
                FileHandler fh = new FileHandler("log.txt");
                fh.setFormatter(new SimpleFormatter());
                logger.addHandler(fh);
                logger.info(welcome);
            } catch (SecurityException e) {
                e.printStackTrace();
            } catch (IOException e) {
                e.printStackTrace();
            }

            DAOFactory factory = new FlatFileDAOFactory(args[0], args[1]);
            Datetime.MAX.setDay(
                    factory.getParametersDAO().getParameters().getNbDays());

            feasible = isFeasible(factory);
            score = calculateCost(factory);

            System.out.println("Feasible : " + (feasible ? "True" : "False"));
            System.out.println("Score 1 : " + score);
        } catch (ROADEFException e) {
            logger.log(Level.SEVERE, e.getError());
        } catch (Exception e) {
            e.printStackTrace();
        }

    }

    private static void printUsage() {
        System.out.println("Usage: java -jar checker.jar path_to_instance/myInstance.zip path_to_solution/mySolution.csv");
    }

    public static boolean isFeasible(DAOFactory factory) throws ROADEFException {
        ConstraintComposite constraint = new ConstraintComposite();
        constraint.add(new SolutionConstraint(factory));
        constraint.add(new EntranceConstraint(factory));
        constraint.add(new ExitConstraint(factory));
        constraint.add(new ResourceUseConstraint(factory));
        constraint.add(new TransitionConstraint(factory));
        constraint.add(new MaintenanceOperationsConstraint(factory));
        constraint.add(new JunctionDisjunctionOperationsConstraint(
                factory));
        constraint.add(new TrackGroupUseConstraint(factory));
        constraint.add(new MinimumReverseTimeConstraint(factory));
        constraint.add(new AtMostOneTrainPerDepartureConstraint(factory));
        constraint.add(new DepartureRequiredDBMConstraint(factory));
        constraint.add(new DepartureRequiredTBMConstraint(factory));
        constraint.add(new DepartureTrainCategoryCompatibilityConstraint(
                factory));
        constraint.add(new AssembledTrainCategoryCompatibilityConstraint(
                factory));
        constraint.add(new ResourceTrainCategoryCompatibilityConstraint(
                factory));
        constraint.add(new PlatformUseMaximumDurationConstraint(factory));
        constraint.add(new ResourceUseMinimumDurationConstraint(factory));
        constraint.add(new ImposedConsumptionConstraint(factory));
        constraint.add(new TrackLengthConstraint(factory));
        constraint.add(new SingleTrackCapacityConstraint(factory));
        constraint.add(new MaintenanceResourcesCapacityConstraint(factory));
        constraint.add(new TrainOrderOnIndividualTrackConstraint(factory));
        constraint.add(new TrainOrderAfterJunctionConstraint(factory));
        constraint.add(new ExitFromResourceAfterDisjunctionConstraint(factory));
        constraint.add(new AssembledTrainsSimultaneousEventConstraint(
                factory));
        constraint.add(new DepartureOrderConstraint(factory));
        constraint.add(new YardOverloadConstraint(factory));
        constraint.add(new TrackGroupConflictConstraint(factory));

        return constraint.validate();
    }

    public static float calculateCost(DAOFactory factory) throws ROADEFException {
        CostComposite f = new CostComposite();

        f.add(new UncoveredDeparturesCost(factory));
        f.add(new OverMaintenanceCost(factory));
        f.add(new JunctionDisjunctionCost(factory));
        f.add(new PlatformUsageCost(factory));
        f.add(new PreferredPlatformAssignementCost(factory));
        f.add(new TrainReuseCost(factory));
        f.add(new UncoveredArrivalCost(factory));

        return f.calculate();
    }

}
