// Copyright 2013 SNCF. All rights reserved.
package com.sncf.checker;

import java.util.Collections;
import java.util.List;
import java.util.ListIterator;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.sncf.checker.bo.ArrDepSequence;
import com.sncf.checker.bo.Departure;
import com.sncf.checker.bo.Event;
import com.sncf.checker.bo.Fill;
import com.sncf.checker.bo.Gate;
import com.sncf.checker.bo.Platform;
import com.sncf.checker.bo.Resource;
import com.sncf.checker.bo.Train;
import com.sncf.checker.common.Datetime;
import com.sncf.checker.common.ROADEFException;
import com.sncf.checker.io.DAOFactory;
import com.sncf.checker.io.DepartureDAO;
import com.sncf.checker.io.TrainDAO;

/**
 * Exit of the system.
 *
 * @author YDS
 *
 */
public class ExitConstraint implements Constraint {

    /** Interface to Data layer. */
    private DAOFactory factory;

    /**
     * Constructor.
     *
     * @param factory dao factory to set
     */
    public ExitConstraint(final DAOFactory factory) {
        this.factory = factory;
    }

    /** {@inheritDoc} */
    public final boolean validate() throws ROADEFException {
        TrainDAO trainDAO = factory.getTrainDAO();
        DepartureDAO departureDAO = factory.getDepartureDAO();
        boolean valid = true;
        Logger logger = Logger.getLogger("logger");

        List<Train> trainList = trainDAO.loadAll();
        for (Train train : trainList) {
            List<Fill> fillList = train.getFill();
            if (fillList.isEmpty()) {
                continue;
            }
            Departure departure = null;
            boolean isDepartureCorrect = true;
            for (Fill fill : fillList) {
                if (fill.getEvent().equals(Event.Departure)) {
                    departure = departureDAO.loadByName(fill.getComplement());
                    if (departure == null) {
                        logger.log(Level.INFO, "Exit of the system error : "
                            + train.getName() + " has an unknown departure"
                            + " (" + fill.getComplement() + ").");
                        isDepartureCorrect = false;
                        valid = false;

                    }
                    break;
                }
            }
            if (!isDepartureCorrect) {
                continue;
            }

            ListIterator<Fill> itr = fillList.listIterator(
                    fillList.size());
            Fill fill = null;
            // train associated with a departure
            if (departure != null)
            {


                ArrDepSequence sequence = null;
                List<ArrDepSequence> arrDepSequenceList =
                        departure.getDepSequence();
                Collections.sort(arrDepSequenceList
                        , new SortArrDepSequenceByOrder());
                ListIterator<ArrDepSequence> itrSequence =
                        arrDepSequenceList.listIterator(
                        arrDepSequenceList.size());

                if (!itr.hasPrevious()) {
                    logger.log(Level.INFO, "Exit of the system error : "
                        + train.getName() + " must end with an ExitSystem "
                        + "event (schedule missing).");
                    valid = false;
                    continue;
                }
                fill = itr.previous();
                sequence = departure.getDepSequence().get(
                        departure.getDepSequence().size() - 1);
                if (!fill.getEvent().equals(Event.ExitSystem)) {
                    logger.log(Level.INFO, "Exit of the system error : "
                        + train.getName() + " must end with an ExitSystem "
                        + "event (" + fill.getEvent().toString() + " found).");
                    valid = false;
                    continue;
                }
                if (fill.getResource() != sequence.getTrackGroup()) {
                    logger.log(Level.INFO, "Exit of the system error : "
                        + train.getName() + " ExitSystem event must occur "
                        + "on " + sequence.getTrackGroup().getName() + " ("
                        + fill.getResource().getName() + " found).");
                    valid = false;
                    continue;
                }
                boolean isSequenceValid = true;
                boolean firstTg = true;
                while (itrSequence.hasPrevious()) {
                    sequence = itrSequence.previous();
                    if (!itr.hasPrevious()) {
                        logger.log(Level.INFO, "Exit of the system error : "
                            + train.getName() + " ExitResource event expected"
                            + " (schedule missing).");
                        valid = false;
                        isSequenceValid = false;
                        break;
                    }
                    fill = itr.previous();
                    if (!fill.getEvent().equals(Event.ExitResource)) {
                        logger.log(Level.INFO, "Exit of the system error : "
                            + train.getName() + " ExitResource expected ("
                            + fill.getEvent().toString() + " at "
                            + fill.getDate().toString() + " found).");
                        valid = false;
                        isSequenceValid = false;
                        break;
                    }
                    if (fill.getResource() != sequence.getTrackGroup()) {
                        logger.log(Level.INFO, "Exit of the system error : "
                            + train.getName() + " ExitResource event must "
                            + "occur on " + sequence.getTrackGroup().getName()
                            + " (" + fill.getResource().getName() + " at "
                            + fill.getDate().toString() + " found).");
                        valid = false;
                        isSequenceValid = false;
                        break;
                    }
                    if (firstTg) {
                        Gate exitGate = null;
                        for (Gate gate : fill.getResource().getGates()) {
                            if (fill.getGate().equals(gate.getSide() + gate.getIndex())) {
                                exitGate = gate;
                                break;
                            }
                        }
                        if (exitGate == null) {
                            logger.log(Level.INFO, "Exit of the system error : "
                                    + train.getName() + " ExitResource event must"
                                    + " define an existing gate on "
                                    + sequence.getTrackGroup().getName());
                            valid = false;
                            isSequenceValid = false;
                            break;
                        }

                        else if (exitGate.getNeighborResource() != null) {
                            logger.log(Level.INFO, "Exit of the system error : "
                                    + train.getName() + " has an incorrect "
                                    + "gate at "
                                    + fill.getResource().getName()
                                    + " (" + fill.getGate() + " is not an entrance of "
                                    + "the system as it is connected with "
                                    + exitGate.getNeighborResource().getName()
                                    + ")");
                            valid = false;
                            isSequenceValid = false;
                            break;
                        }
                    }
                    firstTg = false;

                    if (!itr.hasPrevious()) {
                        logger.log(Level.INFO, "Exit of the system error : "
                            + train.getName() + " EnterResource expected "
                            + "(schedule missing).");
                        valid = false;
                        isSequenceValid = false;
                        break;
                    }

                    fill = itr.previous();
                    if (!fill.getEvent().equals(Event.EnterResource)) {
                        logger.log(Level.INFO, "Exit of the system error : "
                            + train.getName() + " EnterResource expected ("
                            + fill.getEvent().toString() + " at "
                            + fill.getDate().toString() + " found).");
                        valid = false;
                        isSequenceValid = false;
                        break;
                    }
                    if (fill.getResource() != sequence.getTrackGroup()) {
                        logger.log(Level.INFO, "Exit of the system error : "
                            + train.getName() + " EnterResource event "
                            + "must occur on " + sequence.getTrackGroup().getName()
                            + " (" + fill.getResource().getName() + " at "
                            + fill.getDate().toString() + " found).");
                        valid = false;
                        isSequenceValid = false;
                        break;
                    }
                }
                if (!isSequenceValid) {
                    continue;
                }
                if (!itr.hasNext()) {
                    logger.log(Level.INFO, "Exit of the system error : "
                        + train.getName() + " ExitResource on a platform "
                        + "expected (schedule missing).");
                    valid = false;
                    continue;
                }
                fill = itr.previous();
                if (!fill.getEvent().equals(Event.ExitResource)) {
                    logger.log(Level.INFO, "Exit of the system error : "
                        + train.getName() + " ExitResource expected ("
                        + fill.getEvent().toString() + " at "
                        + fill.getDate().toString() + " found).");
                    valid = false;
                    isSequenceValid = false;
                    continue;
                }
                if (fill.getResource().getClass() != Platform.class) {
                    logger.log(Level.INFO, "Exit of the system error : "
                        + train.getName() + " Platform expected ("
                        + fill.getResource().getClass().getName() + " at "
                        + fill.getDate().toString() + " found).");
                    valid = false;
                    continue;
                }
                if (departure.getDepTime().getTime()
                    != fill.getDate().getTime()) {
                    logger.log(Level.INFO, "Exit of the system error : "
                        + train.getName() + " Departure time expected at "
                        + departure.getDepTime().toString() + " ("
                        + fill.getDate().toString() + " found).");
                    valid = false;
                    continue;
                }
                if (!itr.hasPrevious()) {
                    logger.log(Level.INFO, "Exit of the system error : "
                        + train.getName() + " Departure on a platform "
                        + "expected (schedule missing).");
                    valid = false;
                    continue;
                }
                fill = itr.previous();
                if (!fill.getEvent().equals(Event.Departure)) {
                    logger.log(Level.INFO, "Exit of the system error : "
                        + train.getName() + " Departure expected ("
                        + fill.getEvent().toString() + " at "
                        + fill.getDate().toString() + " found).");
                    valid = false;
                    continue;
                }
                if (departure.getDepTime().getTime()
                        != fill.getDate().getTime()) {
                    logger.log(Level.INFO, "Exit of the system error : "
                        + train.getName() + " Departure time expected at "
                        + departure.getDepTime().toString() + " ("
                        + fill.getDate().toString() + " found).");
                    valid = false;
                    continue;
                }
            // train left in the system
            } else {
                Resource exitResource = null;
                if (!itr.hasPrevious()) {
                    logger.log(Level.INFO, "Exit of the system error : "
                        + train.getName() + " must end with an ExitSystem "
                        + "event (schedule missing).");
                    valid = false;
                    continue;
                }
                fill = itr.previous();
                exitResource = fill.getResource();
                if (!fill.getEvent().equals(Event.ExitSystem)) {
                    logger.log(Level.INFO, "Exit of the system error : "
                        + train.getName() + " must end with an ExitSystem "
                        + "event (" + fill.getEvent().toString() + " found).");
                    valid = false;
                    continue;
                }
                if (fill.getDate().getTime() != Datetime.MAX.getTime()) {
                    logger.log(Level.INFO, "Exit of the system error : "
                        + train.getName() + " EnterSystem event must occur "
                        + "occur at " + Datetime.MAX.toString() + " ("
                        + fill.getDate().toString() + " found).");
                    valid = false;
                    continue;
                }
                if (!itr.hasPrevious()) {
                    logger.log(Level.INFO, "Exit of the system error : "
                        + train.getName() + " ExitSystem event must be "
                        + "preceded by an ExitResource event "
                        + "(schedule missing).");
                    valid = false;
                    continue;
                }
                fill = itr.previous();
                if (!fill.getEvent().equals(Event.ExitResource)) {
                    logger.log(Level.INFO, "Exit of the system error : "
                        + train.getName() + " ExitSystem event must be "
                        + "preceded by an ExitResource event ("
                        + fill.getEvent().toString() + " found).");
                    valid = false;
                    continue;
                }
                if (fill.getResource() != exitResource) {
                    logger.log(Level.INFO, "Exit of the system error : "
                        + train.getName() + " last exit event must "
                        + "occur on " + exitResource.getName() + " ("
                        + fill.getResource().getName() + " found).");
                    valid = false;
                    continue;
                }
                if (fill.getDate().getTime() != Datetime.MAX.getTime()) {
                    logger.log(Level.INFO, "Exit of the system error : "
                        + train.getName() + " last ExitRessource event must "
                        + "occur at " + Datetime.MAX.toString() + " ("
                        + fill.getDate().toString() + " found).");
                    valid = false;
                    continue;
                }
            }
            while (itr.hasPrevious()) {
                fill = itr.previous();
                if (fill.getEvent().equals(Event.Departure)) {
                    logger.log(Level.INFO, "Exit of the system error : "
                            + train.getName() + " Departure event unexpected at "
                            + fill.getDate().toString());
                    valid = false;
                }
            }
        }

        return valid;
    }

}
