// Copyright 2013 SNCF. All rights reserved.
package com.sncf.checker;

import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.sncf.checker.bo.Event;
import com.sncf.checker.bo.TrackGroup;
import com.sncf.checker.bo.Fill;
import com.sncf.checker.bo.Train;
import com.sncf.checker.common.Duration;
import com.sncf.checker.common.ROADEFException;
import com.sncf.checker.io.DAOFactory;
import com.sncf.checker.io.TrainDAO;

/**
 * Junction and disjunction operations.
 *
 * @author YDS
 *
 */
public class JunctionDisjunctionOperationsConstraint implements Constraint {

    /** Interface to Data layer. */
    private DAOFactory factory;

    /**
     * Constructor.
     *
     * @param factory dao factory to set
     */
    public JunctionDisjunctionOperationsConstraint(final DAOFactory factory) {
        this.factory = factory;
    }

    /** {@inheritDoc} */
    public final boolean validate() throws ROADEFException {
        TrainDAO trainDAO = factory.getTrainDAO();
        boolean valid = true;
        Logger logger = Logger.getLogger("logger");

        List<Train> trainList = trainDAO.loadAll();
        for (Train train : trainList) {
            IState state = new DefaultState(factory);
            for (Fill fill : train.getFill()) {
                 state = state.handle(fill, logger);
            }
            if (!state.isValid()) {
                valid = false;
            }
        }

        return valid;
    }

    /**
     * state pattern.
     *
     * @author YDS
     *
     */
    private interface IState {

        /**
         * Process a Fill object.
         *
         * @param fill Fill to process
         * @param logger log
         * @return next state
         */
        IState handle(final Fill fill, final Logger logger);
        /**
         * Tells if the process is in a valid state.
         *
         * @return true if valid, false otherwise
         */
        boolean isValid();
    }

    /**
     * default state.
     *
     * @author YDS
     *
     */
    private class DefaultState implements IState {
        /** Validity of the solution. */
        private boolean valid;
        /** Interface to Data layer. */
        private DAOFactory factory;

        /**
         * Constructor.
         *
         * @param factory dao factory to set
         */
        public DefaultState(final DAOFactory factory) {
            valid = true;
            this.factory = factory;
        }

        /**
         * Constructor.
         *
         * @param factory dao factory to set
         * @param valid validity
         */
        public DefaultState(final DAOFactory factory, final boolean valid) {
            this.valid = valid;
            this.factory = factory;
        }

        /** {@inheritDoc} */
        public IState handle(final Fill fill, final Logger logger) {
            if (fill.getEvent().equals(Event.EnterResource)
                    && fill.getResource().getClass() != TrackGroup.class) {
                return new ResourceState(factory, valid);
            } else if (fill.getEvent().equals(Event.BegJunction)) {
                logger.log(Level.INFO, "Junctions and disjunctions operations "
                        + "error : " + fill.getTrain().getName()
                        + " BegJunction event must occur after"
                        + " an EnterResource event on a resource at "
                        + fill.getDate().toString() + ".");
                valid = false;
            } else if (fill.getEvent().equals(Event.BegDisjunction)) {
                logger.log(Level.INFO, "Junctions and disjunctions operations "
                        + "error : " + fill.getTrain().getName()
                        + " BegDisjunction event must occur after "
                        + "an EnterResource event on a resource at "
                        + fill.getDate().toString() + ".");
                valid = false;
            } else if (fill.getEvent().equals(Event.EndJunction)) {
                logger.log(Level.INFO, "Junctions and disjunctions operations "
                        + "error : " + fill.getTrain().getName()
                        + " EndJunction event must occur after "
                        + "a BegJunction event at "
                        + fill.getDate().toString() + ".");
                valid = false;
            } else if (fill.getEvent().equals(Event.EndDisjunction)) {
                logger.log(Level.INFO, "Junctions and disjunctions operations "
                        + "error : " + fill.getTrain().getName()
                        + " EndDisjunction event must occur after a "
                        + "BegDisjunction event at "
                        + fill.getDate().toString() + ".");
                valid = false;
            }
            return this;
        }

        /** {@inheritDoc} */
        public boolean isValid() {
            return valid;
        }
    }

    /**
     * On a resource state.
     *
     * @author YDS
     *
     */
    private class ResourceState implements IState {
        /** Validity of the solution. */
        private boolean valid;
        /** Interface to Data layer. */
        private DAOFactory factory;

        /**
         * Constructor.
         *
         * @param factory dao factory to set
         * @param valid validity
         */
        public ResourceState(final DAOFactory factory, final boolean valid) {
            this.valid = valid;
            this.factory = factory;
        }

        /** {@inheritDoc} */
        public IState handle(final Fill fill, final Logger logger) {
            if (fill.getEvent().equals(Event.BegJunction)) {
                return new JunctionState(factory, fill, valid);
            } else if (fill.getEvent().equals(Event.BegDisjunction)) {
                return new DisjunctionState(factory, fill, valid);
            } else if (fill.getEvent().equals(Event.ExitResource)) {
                return new DefaultState(factory, valid);
            } else if (fill.getEvent().equals(Event.EndJunction)) {
                logger.log(Level.INFO, "Junctions and disjunctions operations "
                        + "error : "
                        + fill.getTrain().getName() + " EndJunction event "
                        + "must occur after a BegJunction event at "
                        + fill.getDate().toString() + ".");
                valid = false;
            } else if (fill.getEvent().equals(Event.EndDisjunction)) {
                logger.log(Level.INFO, "Junctions and disjunctions operations "
                        + "error : " + fill.getTrain().getName()
                        + " EndDisjunction event must occur after a "
                        + "BegDisjunction event on a maintenance facility at "
                        + fill.getDate().toString() + ".");
                valid = false;
            }
            return this;
        }

        /** {@inheritDoc} */
        public boolean isValid() {
            return valid;
        }
    }

    /**
     * Junction operation state.
     *
     * @author YDS
     *
     */
    private class JunctionState implements IState {
        /** Junction beginning. */
        private Fill begin;
        /** Validity of the solution. */
        private boolean valid;
        /** Interface to Data layer. */
        private DAOFactory factory;

        /**
         * Constructor.
         *
         * @param factory dao factory to set
         * @param fill junction beginning
         * @param valid validity
         */
        public JunctionState(final DAOFactory factory
                , final Fill fill, final boolean valid) {
            begin = fill;
            this.valid = valid;
            this.factory = factory;
        }

        /** {@inheritDoc} */
        public IState handle(final Fill fill, final Logger logger) {
            if (fill.getEvent().equals(Event.EndJunction)) {
                int duration = fill.getDate().getTime()
                        - begin.getDate().getTime();
                Duration junTime = factory.getParametersDAO()
                        .getParameters().getJunTime();
                if (duration != junTime.getTime()) {
                    logger.log(Level.INFO, "Junctions and disjunctions "
                            + "operations error : " + fill.getTrain().getName()
                            + " junction time at "
                            + fill.getDate().toString() + " should be "
                            + junTime.toString()
                            + " (" + new Duration(duration).toString()
                            + " found).");
                    valid = false;
                }
            } else {
                logger.log(Level.INFO, "Junctions and disjunctions operations "
                        + "error : " + fill.getTrain().getName()
                        + " EndJunction event was expected "
                        + "at " + fill.getDate().toString() + ".");
                valid = false;
            }
            return new ResourceState(factory, valid);
        }

        /** {@inheritDoc} */
        public boolean isValid() {
            return valid;
        }
    }

    /**
     * Disjunction operation state.
     *
     * @author YDS
     *
     */
    private class DisjunctionState implements IState {
        /** Disjunction beginning. */
        private Fill begin;
        /** Validity of the solution. */
        private boolean valid;
        /** Interface to Data layer. */
        private DAOFactory factory;

        /**
         * Constructor.
         *
         * @param factory dao factory to set
         * @param fill disjunction beginning
         * @param valid validity
         */
        public DisjunctionState(final DAOFactory factory, final Fill fill
                , final boolean valid) {
            begin = fill;
            this.valid = valid;
            this.factory = factory;
        }

        /** {@inheritDoc} */
        public IState handle(final Fill fill, final Logger logger) {
            if (fill.getEvent().equals(Event.EndDisjunction)) {
                int duration = fill.getDate().getTime()
                        - begin.getDate().getTime();
                Duration junTime = factory.getParametersDAO().getParameters()
                        .getDisjTime();
                if (duration != junTime.getTime()) {
                    logger.log(Level.INFO, "Junctions and disjunctions "
                            + "operations error : " + fill.getTrain().getName()
                            + " disjunction time at "
                            + fill.getDate().toString() + " should be "
                            + junTime.toString() + " ("
                            + new Duration(duration).toString() + " found).");
                    valid = false;
                }
            } else {
                logger.log(Level.INFO, "Junctions and disjunctions operations "
                		+ "error : " + fill.getTrain().getName()
                		+ " EndDisjunction event was expected "
                        + "at " + fill.getDate().toString() + ".");
                valid = false;
            }
            return new ResourceState(factory, valid);
        }

        /** {@inheritDoc} */
        public boolean isValid() {
            return valid;
        }
    }
}
