// Copyright 2013 SNCF. All rights reserved.
package com.sncf.checker;

import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.sncf.checker.bo.Event;
import com.sncf.checker.bo.Fill;
import com.sncf.checker.bo.TrackGroup;
import com.sncf.checker.bo.Train;
import com.sncf.checker.common.Datetime;
import com.sncf.checker.common.Duration;
import com.sncf.checker.common.ROADEFException;
import com.sncf.checker.io.DAOFactory;
import com.sncf.checker.io.ParametersDAO;
import com.sncf.checker.io.TrainDAO;

/**
 * Minimum reverse time.
 *
 * @author YDS
 *
 */
public class MinimumReverseTimeConstraint implements Constraint {

    /** Interface to Data layer. */
    private DAOFactory factory;

    /**
     * Constructor.
     *
     * @param factory dao factory to set
     */
    public MinimumReverseTimeConstraint(final DAOFactory factory) {
        this.factory = factory;
    }

    /** {@inheritDoc} */
    public boolean validate() throws ROADEFException {
        TrainDAO trainDAO = factory.getTrainDAO();
        ParametersDAO parametersDAO = factory.getParametersDAO();
        Duration revTime = parametersDAO.getParameters().getRevTime();
        boolean valid = true;
        Logger logger = Logger.getLogger("logger");

        List<Train> trainList = trainDAO.loadAll();
        for (Train train : trainList) {
            IState state = new DefaultState(revTime);
            for (Fill fill : train.getFill()) {
                 state = state.handle(fill, logger);
            }
            if (!state.isValid()) {
                valid = false;
            }
        }

        return valid;
    }

    /**
     * state pattern.
     *
     * @author YDS
     *
     */
    private interface IState {
        /**
         * Process a Fill object.
         *
         * @param fill Fill to process
         * @param logger log
         * @return next state
         */
        IState handle(final Fill fill, final Logger logger);
        /**
         * Tells if the process is in a valid state.
         *
         * @return true if valid, false otherwise
         */
        boolean isValid();
    }

    /**
     * default state.
     *
     * @author YDS
     *
     */
    private class DefaultState implements IState {
        /** Solution validity. */
        private boolean valid;
        /** minimum reverse time. */
        Duration revTime;

        /**
         * Constructor.
         *
         * @param revTime minimum reverse time
         */
        public DefaultState(final Duration revTime) {
            valid = true;
            this.revTime = revTime;
        }

        /**
         * Constructor.
         *
         * @param revTime minimum reverse time
         * @param valid validity
         */
        public DefaultState(final Duration revTime, final boolean valid) {
            this.valid = valid;
            this.revTime = revTime;
        }

        /** {@inheritDoc} */
        public IState handle(final Fill fill, final Logger logger)
        {
            if (fill.getEvent().equals(Event.EnterResource)
                    && fill.getResource().getClass() != TrackGroup.class
                    && fill.getDate().getTime() != Datetime.MIN.getTime()) {
                return new ReversibleState(revTime, fill, valid);
            }
            return this;
        }

        /** {@inheritDoc} */
        public boolean isValid() {
            return valid;
        }
    }

    /**
     * Train can be reversed.
     *
     * @author YDS
     *
     */
    private class ReversibleState implements IState {
        /** Entering resource time. */
        private Fill begin;
        /** Solution validity. */
        private boolean valid;
        /** minimum reverse time. */
        private Duration revTime;

        /**
         * Constructor.
         *
         * @param revTime minimum reverse time
         * @param fill entering resource time
         * @param valid validity
         */
        public ReversibleState(final Duration revTime, final Fill fill
                , final boolean valid) {
            begin = fill;
            this.valid = valid;
            this.revTime = revTime;
        }

        /** {@inheritDoc} */
        public IState handle(final Fill fill, final Logger logger)
        {
            if (fill.getEvent().equals(Event.ExitResource)) {
                if (!(begin.getGate().isEmpty() || fill.getGate().isEmpty())) {
                    if (begin.getGate().charAt(0)
                            == fill.getGate().charAt(0)) {
                        int duration = fill.getDate().getTime()
                                - begin.getDate().getTime();
                        if (duration < revTime.getTime()
                                && fill.getDate().getTime()
                                != Datetime.MAX.getTime()) {
                            logger.log(Level.INFO, "Minimum reverse time error : "
                                    + fill.getTrain().getName()
                                    + " minimum reverse time on "
                                    + begin.getResource().getName()
                                    + " at " + begin.getDate().toString()
                                    + " shoud be " + revTime.toString() + " ("
                                    + new Duration(duration).toString()
                                    + " found).");
                            valid = false;
                        }
                    }
                    return new DefaultState(revTime, valid);
                }
            }
            return this;
        }

        /** {@inheritDoc} */
        public boolean isValid() {
            return valid;
        }
    }
}