// Copyright 2013 SNCF. All rights reserved.
package com.sncf.checker;

import java.util.HashMap;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.sncf.checker.bo.Arrival;
import com.sncf.checker.bo.Departure;
import com.sncf.checker.bo.Event;
import com.sncf.checker.bo.Fill;
import com.sncf.checker.bo.Parameters;
import com.sncf.checker.bo.Train;
import com.sncf.checker.common.Duration;
import com.sncf.checker.common.ROADEFException;
import com.sncf.checker.io.DAOFactory;
import com.sncf.checker.io.DepartureDAO;
import com.sncf.checker.io.TrainDAO;

/**
 * Over-maintenance cost.
 *
 * @author YDS
 *
 */
public class OverMaintenanceCost implements Cost {

    /** Interface to Data layer. */
    private final DAOFactory factory;

    /**
     * Constructor.
     *
     * @param factory dao factory to set
     */
    public OverMaintenanceCost(final DAOFactory factory) {
        this.factory = factory;
    }

    /** {@inheritDoc} */
    public float calculate() throws ROADEFException {
        HashMap<Departure, Train> assigned = new HashMap<Departure, Train>();
        TrainDAO dao = factory.getTrainDAO();
        DepartureDAO departureDAO = factory.getDepartureDAO();
        Parameters parameters = factory.getParametersDAO().getParameters();
        int sumRemDBM = 0;
        int sumRemTBM = 0;
        Logger logger = Logger.getLogger("logger");

        List<Train> trainList = dao.loadAll();
        for (Train train : trainList) {
            for (Fill fill : train.getFill()) {
                if (fill.getEvent().equals(Event.Departure)) {
                    Departure dep = departureDAO.loadByName(
                            fill.getComplement());
                    if (dep != null) {
                        assigned.put(dep, train);
                    }
                }
            }
        }

        for (Train train : trainList) {
            int remDBM = train.getRemDBM();
            Departure departure = null;
            Duration remTBM = train.getRemTBM();
            Arrival arrival = train.getArrival();
            if (arrival != null) {
                Departure linkedDep = arrival.getLinkedDep();
                if (linkedDep != null) {
                    Train prevTrain = assigned.get(linkedDep);
                    if (prevTrain != null) {
                        continue;
                    }
                }
            }

            for (Fill fill : train.getFill()) {
                if (fill.getEvent().equals(Event.BegMaintenance)) {
                    if (fill.getComplement().equals("D")) {
                        logger.log(Level.INFO, "Over maintenance cost : "
                                + train.getName() + " has " + remDBM
                                + " km remaining at " + fill.getDate().toString()
                                + " maintenance. Score += "
                                + parameters.getRemDCost() + " * "
                                + remDBM + " km = "
                                + parameters.getRemDCost() * remDBM);
                        sumRemDBM += remDBM;
                        remDBM = train.getCategory().getMaxDBM();
                    } else if (fill.getComplement().equals("T")) {
                        logger.log(Level.INFO, "Over maintenance cost : "
                                + train.getName() + " has " + remTBM
                                + " remaining at " + fill.getDate().toString()
                                + " maintenance. Score += "
                                + parameters.getRemTCost() + " * "
                                + remTBM.getTime() + " s = "
                                + parameters.getRemTCost() * remTBM.getTime());
                        sumRemTBM += remTBM.getTime();
                        remTBM = train.getCategory().getMaxTBM();
                    }
                } else if (fill.getEvent().equals(Event.Departure)) {
                    departure = departureDAO.loadByName(fill.getComplement());
                }
            }

            while (departure != null) {
                remDBM = remDBM - departure.getReqDBM();
                int remTBMsec = remTBM.getTime() - departure.getReqTBM().getTime();
                remTBM = new Duration(remTBMsec);
                if (departure.getLinkedArr() != null) {
                    train = departure.getLinkedArr().getTrain();
                    departure = null;
                    for (Fill fill : train.getFill()) {
                        if (fill.getEvent().equals(Event.BegMaintenance)) {
                            if (fill.getComplement().equals("D")) {
                                logger.log(Level.INFO, "Over maintenance cost : "
                                        + train.getName() + " has " + remDBM
                                        + " km remaining at " + fill.getDate().toString()
                                        + " maintenance. Score += "
                                        + parameters.getRemDCost() + " * "
                                        + remDBM + " km = "
                                        + parameters.getRemDCost() * remDBM);
                                sumRemDBM += remDBM;
                                remDBM = train.getCategory().getMaxDBM();
                            } else if (fill.getComplement().equals("T")) {
                                logger.log(Level.INFO, "Over maintenance cost : "
                                        + train.getName() + " has " + remTBM
                                        + " remaining at " + fill.getDate().toString()
                                        + " maintenance. Score += "
                                        + parameters.getRemTCost() + " * "
                                        + remTBM.getTime() + " s = "
                                        + parameters.getRemTCost() * remTBM.getTime());
                                sumRemTBM += remTBM.getTime();
                                remTBM = train.getCategory().getMaxTBM();
                            }
                        } else if (fill.getEvent().equals(Event.Departure)) {
                            departure = departureDAO.loadByName(
                                    fill.getComplement());
                        }
                    }
                } else {
                    departure = null;
                }
            }
        }
        return sumRemDBM * parameters.getRemDCost() + sumRemTBM
                * parameters.getRemTCost();
    }

}
