// Copyright 2013 SNCF. All rights reserved.
package com.sncf.checker;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.sncf.checker.bo.Arrival;
import com.sncf.checker.bo.Event;
import com.sncf.checker.bo.Fill;
import com.sncf.checker.bo.ImposedConsumption;
import com.sncf.checker.bo.JointArrDep;
import com.sncf.checker.bo.JointArrDepEntry;
import com.sncf.checker.bo.TrackGroup;
import com.sncf.checker.bo.Train;
import com.sncf.checker.common.Datetime;
import com.sncf.checker.common.ROADEFException;
import com.sncf.checker.io.ArrivalDAO;
import com.sncf.checker.io.DAOFactory;
import com.sncf.checker.io.ImposedConsumptionDAO;
import com.sncf.checker.io.TrackGroupDAO;

public class TrackGroupConflictConstraint implements Constraint {

    private final DAOFactory factory;

    public TrackGroupConflictConstraint(DAOFactory factory) {
        this.factory = factory;
    }

    public boolean validate() throws ROADEFException {
        boolean valid = true;
        TrackGroupDAO trackGroupDAO = factory.getTrackGroupDAO();
        Logger logger = Logger.getLogger("logger");

        for (TrackGroup trackGroup : trackGroupDAO.loadAll()) {
            List<Fill> fillList = trackGroup.getFill();
            Collections.sort(fillList, new SortFillByDate());
            LinkedList<Fill> simultaneous = new LinkedList<Fill>();
            ArrayList<Move> moveList = new ArrayList<Move>();

            // imposed consumptions
            ImposedConsumptionDAO consumptionDAO =
                    factory.getImposedConsumptionDAO();
            for (ImposedConsumption consumption : consumptionDAO.loadAll()) {
                if (consumption.getResource() == trackGroup) {
                    Move move = createMove(consumption);
                    if (move.isValid()) {
                        moveList.add(move);
                    }
                }
            }

            for (int i = 0; i < fillList.size(); i++) {
                Fill beg = fillList.get(i);
                Fill end = null;
                if (beg.getEvent().equals(Event.EnterResource)) {
                    if (simultaneous.isEmpty()) {
                        simultaneous.add(beg);
                    } else {
                        if (simultaneous.getFirst().getDate().getTime() < beg.getDate().getTime()) {
                            simultaneous.clear();
                            simultaneous.add(beg);
                        } else if (isAssembled(beg, simultaneous)) {
                            continue;
                        } else {
                            simultaneous.add(beg);
                        }
                    }
                    for (int j = i + 1; j < fillList.size(); j++) {
                        Fill fill = fillList.get(j);
                        if (fill.getTrain() == beg.getTrain()) {
                            if (fill.getEvent().equals(Event.ExitResource)) {
                                end = fill;
                                break;
                            }
                        }
                    }
                    Move move = createMove(beg, end);
                    if (move.isValid()) {
                        moveList.add(move);
                    }
                }

            }

            Collections.sort(moveList);

            int window = trackGroup.getHeadway().getTime() + trackGroup.getTravelTime().getTime();
            for (int i = 0; i < moveList.size(); i++) {
                for (int j = i + 1; j < moveList.size(); j++) {
                    Move m1 = moveList.get(i);
                    Move m2 = moveList.get(j);
                    if ((m2.getDate().getTime() - m1.getDate().getTime()) > window) {
                        break;
                    }
                    if (detectConflict(m1, m2, trackGroup)) {
                        logger.log(Level.INFO, "Track group conflict constraint : "
                                + m1.getTrain() + " (" + m1.getDate().toString()
                                + " " + m1.getOriginSide() + m1.getOriginInd()
                                + " - " +  m1.getDestinationSide() + m1.getDestinationInd()
                                + ") is in conflict with "
                                + m2.getTrain() + " (" + m2.getDate().toString()
                                + " " + m2.getOriginSide() + m2.getOriginInd()
                                + " - " + m2.getDestinationSide() + m2.getDestinationInd()
                                + ") at " + trackGroup.getName() + ".");
                        valid = false;
                    }
                }
            }
        }
        return valid;
    }

    private boolean isAssembled(Fill beg, List<Fill> fillList) throws ROADEFException {
        ArrivalDAO arrivalDAO = factory.getArrivalDAO();
        ArrayList<String> assembledTrains = new ArrayList<String>();
        Train train = beg.getTrain();
        Arrival arrival = train.getArrival();
        if (arrival != null) {
            JointArrDep joint = arrival.getJointArrDep();
            if (joint != null) {
                for (JointArrDepEntry entry : joint.getEntryList()) {
                    Arrival jointArrival = arrivalDAO.loadByName(entry.getArrDep());
                    if (!jointArrival.getTrain().getFill().isEmpty()) {
                        assembledTrains.add(jointArrival.getTrain().getName());
                    }
                }
            } else {
                assembledTrains.add(train.getName());
            }
        } else {
            assembledTrains.add(train.getName());
        }

        for (Fill fill : train.getFill()) {
            if (fill == beg) {
                break;
            }
            if (fill.getEvent().equals(Event.EndJunction) || fill.getEvent().equals(Event.EndDisjunction)) {
                assembledTrains = new ArrayList<String>(Arrays.asList(fill.getComplement().split("\\+")));
            }
        }

        for (Fill fill : fillList) {
            if (assembledTrains.contains(fill.getTrain().getName())) {
                return true;
            }
        }
        return false;
    }

    private boolean detectConflict(Move m1, Move m2, TrackGroup trackGroup) {
        // same direction trains
        if (m1.getOriginSide().equals(m2.getOriginSide())) {
            if (m1.getOriginInd() < m2.getOriginInd()) {
                if (m1.getDestinationInd() >= m2.getDestinationInd()) {
                    if (Math.abs(m1.getDate().getTime() - m2.getDate().getTime()) < trackGroup.getHeadway().getTime()) {
                        return true;
                    }
                }
            } else if (m1.getOriginInd() > m2.getOriginInd()) {
                if (m1.getDestinationInd() <= m2.getDestinationInd()) {
                    if (Math.abs(m1.getDate().getTime() - m2.getDate().getTime()) < trackGroup.getHeadway().getTime()) {
                        return true;
                    }
                }
            } else {
                if (Math.abs(m1.getDate().getTime() - m2.getDate().getTime()) < trackGroup.getHeadway().getTime()) {
                    return true;
                }
            }
        // opposite directions trains
        } else {
            if (m1.getOriginInd() < m2.getDestinationInd()) {
                if (m1.getDestinationInd() >= m2.getOriginInd()) {
                    if (Math.abs(m1.getDate().getTime() - m2.getDate().getTime()) < (trackGroup.getTravelTime().getTime() + trackGroup.getHeadway().getTime())) {
                        return true;
                    }
                }
            } else if (m1.getOriginInd() > m2.getDestinationInd()) {
                if (m1.getDestinationInd() <= m2.getOriginInd()) {
                    if (Math.abs(m1.getDate().getTime() - m2.getDate().getTime()) < (trackGroup.getTravelTime().getTime() + trackGroup.getHeadway().getTime())) {
                        return true;
                    }
                }
            } else {
                if (Math.abs(m1.getDate().getTime() - m2.getDate().getTime()) < (trackGroup.getTravelTime().getTime() + trackGroup.getHeadway().getTime())) {
                    return true;
                }
            }
        }
        return false;
    }

    private Move createMove(Fill beg, Fill end) {
        Move move = new Move();
        if (beg != null) {
            move.setDate(beg.getDate());
            move.setTrain(beg.getTrain().getName());
            if (beg.getGate().length() >= 2) {
                move.setOriginSide(beg.getGate().substring(0, 1));
                try {
                    move.setOriginInd(Integer.parseInt(beg.getGate().substring(1)));
                } catch (Exception e) {
                }
            }
        }
        if (end != null) {
            if (end.getGate().length() >= 2) {
                move.setDestinationSide(end.getGate().substring(0, 1));
                try {
                    move.setDestinationInd(Integer.parseInt(end.getGate().substring(1)));
                } catch (Exception e) {
                }
            }
        }
        return move;
    }

    private Move createMove(ImposedConsumption consumption) {
        Move move = new Move();
        move.setTrain("Imposed consumption");
        move.setDate(consumption.getEntranceTime());
        if (consumption.getOriginGate().length() >= 2) {
            move.setOriginSide(consumption.getOriginGate().substring(0, 1));
            try {
                move.setOriginInd(Integer.parseInt
                        (consumption.getOriginGate().substring(1)));
            } catch (Exception e) {
            }
        }
        if (consumption.getDestinationGate().length() >= 2) {
            move.setDestinationSide(
                    consumption.getDestinationGate().substring(0, 1));
            try {
                move.setDestinationInd(Integer.parseInt(
                        consumption.getDestinationGate().substring(1)));
            } catch (Exception e) {
            }
        }
        return move;
    }

    private class Move implements Comparable<Move> {

        public Move() {
        }

        private String train;

        public String getTrain() {
            return train;
        }

        public void setTrain(String train) {
            this.train = train;
        }

        private Datetime date;

        public Datetime getDate() {
            return date;
        }

        public void setDate(Datetime date) {
            this.date = date;
        }

        String originSide;

        public String getOriginSide() {
            return originSide;
        }

        public void setOriginSide(String originSide) {
            this.originSide = originSide;
        }

        Integer originInd;

        public Integer getOriginInd() {
            return originInd;
        }

        public void setOriginInd(Integer originInd) {
            this.originInd = originInd;
        }

        String destinationSide;

        public String getDestinationSide() {
            return destinationSide;
        }

        public void setDestinationSide(String destinationSide) {
            this.destinationSide = destinationSide;
        }

        Integer destinationInd;

        public Integer getDestinationInd() {
            return destinationInd;
        }

        public void setDestinationInd(Integer destinationInd) {
            this.destinationInd = destinationInd;
        }

        public boolean isValid() {
            if (date == null
                    || originSide == null
                    || originInd == null
                    || destinationSide == null
                    || destinationInd == null) {
                return false;
            } else {
                if (!(originSide.equals("A") || originSide.equals("B"))
                        || !(destinationSide.equals("A") || destinationSide.equals("B"))) {
                    return false;
                }
            }
            return true;
        }

        public int compareTo(Move arg0) {
            if (date.getTime() < arg0.date.getTime()) {
                return -1;
            } else if (date.getTime() > arg0.date.getTime()) {
                return 1;
            } else {
                return 0;
            }

        }
    }
}
