// Copyright 2013 SNCF. All rights reserved.
package com.sncf.checker.common;

/**
 * Application date.
 *
 * @author YDS
 *
 */
public final class Datetime {

    /** number of seconds per minute. */
    public static final int SECONDS_PER_MINUTE = 60;
    /** number of seconds per hour. */
    public static final int SECONDS_PER_HOUR = 3600;
    /** number of seconds per day. */
    public static final int SECONDS_PER_DAY = 86400;
    /** number of minutes per hour. */
    public static final int MINUTES_PER_HOUR = 60;
    /** number of hours per day. */
    public static final int HOURS_PER_DAY = 24;
    /** default day. */
    public static final int DEFAULT_DAY = 1;
    /** default hour. */
    public static final int DEFAULT_HOUR = 0;
    /** default minute. */
    public static final int DEFAULT_MINUTE = 0;
    /** default second. */
    public static final int DEFAULT_SECOND = 0;
    /** minimum Datetime. */
    public static final Datetime MIN = new Datetime("d1 00:00:00");
    /** maximum Datetime. */
    public static final Datetime MAX = new Datetime("d7 23:59:59");

    /**
     * Construct a new date at "d1 00:00".
     */
    public Datetime() {
        day = DEFAULT_DAY;
        hour = DEFAULT_HOUR;
        minute = DEFAULT_MINUTE;
        second = DEFAULT_SECOND;
    }

    /**
     * Construct a new date from a string (di hh:mm:ss). Format di hh:mm is
     *
     * @param date a String representing the date
     */
    public Datetime(final String date) {
        String temp = date.concat("$");
        IState state = new InitialState(this);
        for (int i = 0; i < temp.length(); i++) {
            state = state.handle(Character.toString(temp.charAt(i)));
        }
    }

    /**
     * Construct a new date from a string (di hh:mm:ss). Format di hh:mm is
     *
     *
     * @param date a String representing the date
     */
    public Datetime(int date) {
        int rest = date % SECONDS_PER_MINUTE;
        setSecond(rest);
        date = date - rest;
        rest = date % SECONDS_PER_HOUR;
        setMinute(rest / SECONDS_PER_MINUTE);
        date = date - rest;
        rest = date % SECONDS_PER_DAY;
        setHour(rest / SECONDS_PER_HOUR);
        date = date - rest;
        setDay(date / SECONDS_PER_DAY + 1);
    }

    /** day of the week. */
    private int day;

    /**
     * Get the day.
     *
     * @return day an integer from 1 to 7
     */
    public int getDay() {
        return day;
    }

    /**
     * Set the day.
     *
     * @param day an integer from 1 to 7
     */
    public void setDay(final int day) {
        this.day = day;
    }

    /** hour. */
    private int hour;

    /**
     * Get the hour.
     *
     * @return hour an integer from 0 to 23
     */
    public int getHour() {
        return hour;
    }

    /**
     * Set the hour.
     *
     * @param hour an integer from 0 to 23
     */
    public void setHour(final int hour) {
        if (hour < 0 || hour >= HOURS_PER_DAY) {
            throw new IllegalArgumentException("Number of hours for a datetime should be between 0 an 23");
        }
        this.hour = hour;
    }

    /** minute. */
    private int minute;

    /**
     * Get the minute.
     *
     * @return minute an integer from 0 to 59
     */
    public int getMinute() {
        return minute;
    }

    /**
     * Set the minute.
     *
     * @param minute an integer from 0 to 59
     */
    public void setMinute(final int minute) {
        if (minute < 0 || minute >= MINUTES_PER_HOUR) {
            throw new IllegalArgumentException("Number of minutes for a datetime should be between 0 an 59");
        }
        this.minute = minute;
    }

    /** second. */
    private int second;

    /**
     * Get the second.
     *
     * @return minute an integer from 0 to 59
     */
    public int getSecond() {
        return second;
    }

    public Datetime add(Duration duration) {
        return new Datetime(getTime() + duration.getTime());
    }

    public Datetime substract(Duration duration) {
        return new Datetime(getTime() - duration.getTime());
    }

    /**
     * Set the second.
     *
     * @param second an integer from 0 to 59
     */
    public void setSecond(final int second) {
        if (second < 0 || second >= SECONDS_PER_MINUTE) {
            throw new IllegalArgumentException("Number of seconds for a datetime should be between 0 an 59");
        }
        this.second = second;
    }

    /**
     * Convert the date in seconds since "d1 00:00:00".
     *
     * @return minute
     */
    public int getTime() {
        return (day - 1) * SECONDS_PER_DAY
                + hour * SECONDS_PER_HOUR
                + minute * SECONDS_PER_MINUTE
                + second;
    }

    /**
     * Convert the date in a string representation.
     *
     * @return minute
     */
    @Override
    public String toString() {
        return "d" + String.valueOf(day) + " "
                + (hour < 10 ? "0" : "") + String.valueOf(hour) + ":"
                + (minute < 10 ? "0" : "") + String.valueOf(minute) + ":"
                + (second < 10 ? "0" : "") + String.valueOf(second);
    }

    /**
     * State pattern for parsing string formated datetime.
     *
     * @author YDS
     *
     */
    private interface IState {
        /**
         * Parse the next token.
         *
         * @param c current token
         * @return next state
         */
        IState handle(final String c);
    }

    /**
     * Initial state.
     *
     * @author YDS
     *
     */
    private class InitialState implements IState {
        /** date. */
        private final Datetime date;

        /**
         * Constructor.
         *
         * @param date target Datetime object
         */
        public InitialState(final Datetime date) {
            this.date = date;
        }

        /**
         * Parse the next token.
         *
         * @param c current token
         * @return next state
         */
        public IState handle(final String c) {
            if (c.matches("d")) {
                return new DayState(date);
            } else {
                throw new IllegalArgumentException();
            }
        }
    }

    /**
     * Day parsing state.
     *
     * @author YDS
     *
     */
    private class DayState implements IState {
        /** date. */
        private final Datetime date;
        /** string representation of the day. */
        private String val = "";

        /**
         * Constructor.
         *
         * @param date target Datetime object
         */
        public DayState(final Datetime date) {
            this.date = date;
        }

        /**
         * Parse the next token.
         *
         * @param c current token
         * @return next state
         */
        public IState handle(final String c) {
            if (c.matches("[0-9]")) {
                val = val.concat(c);
                return this;
            } else if (c.matches(" ")) {
                date.setDay(Integer.parseInt(val));
                return new HourState(date);
            } else {
                throw new IllegalArgumentException();
            }
        }
    }

    /**
     * Hour parsing state.
     *
     * @author YDS
     *
     */
    private class HourState implements IState {
        /** date. */
        private final Datetime date;
        /** string representation of the hours. */
        private String val = "";

        /**
         * Constructor.
         *
         * @param date target Datetime object
         */
        public HourState(final Datetime date) {
            this.date = date;
        }

        /**
         * Parse the next token.
         *
         * @param c current token
         * @return next state
         */
        public IState handle(final String c) {
            if (c.matches("[0-9]")) {
                val = val.concat(c);
                return this;
            } else if (c.matches(":")) {
                date.setHour(Integer.parseInt(val));
                return new MinuteState(date);
            } else {
                throw new IllegalArgumentException();
            }
        }
    }

    /**
     * Minute parsing state.
     *
     * @author YDS
     *
     */
    private class MinuteState implements IState {
        /** date. */
        private final Datetime date;
        /** string representation of the minutes. */
        private String val = "";

        /**
         * Constructor.
         *
         * @param date target Datetime object
         */
        public MinuteState(final Datetime date) {
            this.date = date;
        }

        /**
         * Parse the next token.
         *
         * @param c current token
         * @return next state
         */
        public IState handle(final String c) {
            if (c.matches("[0-9]")) {
                val = val.concat(c);
                return this;
            } else if (c.matches("\\$")) {
                date.setMinute(Integer.parseInt(val));
                return new EndState();
            } else if (c.matches(":")) {
                date.setMinute(Integer.parseInt(val));
                return new SecondState(date);
            } else {
                throw new IllegalArgumentException();
            }
        }
    }

    /**
     * Second parsing state.
     *
     * @author YDS
     *
     */
    private class SecondState implements IState {
        /** date. */
        private final Datetime date;
        /** string representation of the seconds. */
        private String val = "";

        /**
         * Constructor.
         *
         * @param date target Datetime object
         */
        public SecondState(final Datetime date) {
            this.date = date;
        }

        /**
         * Parse the next token.
         *
         * @param c current token
         * @return next state
         */
        public IState handle(final String c) {
            if (c.matches("[0-9]")) {
                val = val.concat(c);
                return this;
            } else if (c.matches("\\$")) {
                date.setSecond(Integer.parseInt(val));
                return new EndState();
            } else {
                throw new IllegalArgumentException();
            }
        }
    }

    /**
     * Initial state.
     *
     * @author YDS
     *
     */
    private class EndState implements IState {

        /**
         * Parse the next token.
         *
         * @param c current token
         * @return next state
         */
        public IState handle(final String c) {
            throw new IllegalArgumentException();
        }
    }
}


