// Copyright 2013 SNCF. All rights reserved.
package com.sncf.checker.common;

/**
 * Application date.
 *
 * @author YDS
 *
 */
public final class Duration {

    /** number of seconds per minute. */
    public static final int SECONDS_PER_MINUTE = 60;
    /** number of minutes per hour. */
    public static final int MINUTES_PER_HOUR = 60;
    /** number of seconds per day. */
    public static final int DEFAULT_HOUR = 0;
    /** number of seconds per day. */
    public static final int DEFAULT_MINUTE = 0;
    /** number of seconds per day. */
    public static final int DEFAULT_SECOND = 0;

    /**
     * Construct a new date at "d1 00:00".
     */
    public Duration() {
        hour = DEFAULT_HOUR;
        minute = DEFAULT_MINUTE;
        second = DEFAULT_SECOND;
    }

    /**
     * Construct a new date from a string (di hh:mm:ss). Format di hh:mm is
     *
     *
     * @param date a String representing the date
     */
    public Duration(final String time) {
        String temp = time.concat("$");
        IState state = new HourState(this);
        for (int i = 0; i < temp.length(); i++) {
            state = state.handle(Character.toString(temp.charAt(i)));
        }
    }

    /**
     * Construct a new duration from an integer
     *
     * @param date an integer representing the duration
     */
    public Duration(int date) {
        int rest = date % Datetime.SECONDS_PER_MINUTE;
        setSecond(rest);
        date = date - rest;
        rest = date % Datetime.SECONDS_PER_HOUR;
        setMinute(rest / Datetime.SECONDS_PER_MINUTE);
        date = date - rest;
        setHour(date / Datetime.SECONDS_PER_HOUR);
    }

    /** hour. */
    private int hour;

    /**
     * Get the hour.
     *
     * @return hour an integer
     */
    public int getHour() {
        return hour;
    }

    /**
     * Set the hour.
     *
     * @param hour an integer
     */
    public void setHour(final int hour) {
        this.hour = hour;
    }

    /** minute. */
    private int minute;

    /**
     * Get the minute.
     *
     * @return minute an integer from 1 to 60
     */
    public int getMinute() {
        return minute;
    }

    /**
     * Set the minute.
     *
     * @param minute an integer from 1 to 60
     */
    public void setMinute(final int minute) {
        if (minute < 0 || minute >= MINUTES_PER_HOUR) {
            throw new IllegalArgumentException("Number of minutes for a duration should be between 0 an 59");
        }
        this.minute = minute;
    }

    /** second. */
    private int second;

    /**
     * Get the second.
     *
     * @return minute an integer from 1 to 60
     */
    public int getSecond() {
        return second;
    }

    /**
     * Set the second.
     *
     * @param second an integer from 1 to 60
     */
    public void setSecond(final int second) {
        if (second < 0 || second >= SECONDS_PER_MINUTE) {
            throw new IllegalArgumentException("Number of seconds for a duration should be between 0 an 59");
        }
        this.second = second;
    }

    /**
     * Convert the duration in seconds
     *
     * @return minute
     */
    public int getTime() {
        return hour * Datetime.SECONDS_PER_HOUR
                + minute * Datetime.SECONDS_PER_MINUTE
                + second;
    }

    public Duration add(Duration duration) {
        return new Duration(getTime() + duration.getTime());
    }

    public Duration substract(Duration duration) {
        return new Duration(getTime() - duration.getTime());
    }

    /**
     * Convert the date in a string representation.
     *
     * @return minute
     */
    @Override
    public String toString() {
        return String.valueOf(hour) + ":"
                + (minute < 10 ? "0" : "") + String.valueOf(minute) + ":"
                + (second < 10 ? "0" : "") + String.valueOf(second);
    }

    /**
     * State pattern for parsing string formated duration.
     *
     * @author YDS
     *
     */
    private interface IState {
        /**
         * Parse the next token.
         *
         * @param c current token
         * @return next state
         */
        IState handle(final String c);
    }

    /**
     * Hour parsing state.
     *
     * @author YDS
     *
     */
    private class HourState implements IState {
        /** date. */
        private final Duration date;
        /** string representation of the hours. */
        private String val = "";

        /**
         * Constructor.
         *
         * @param date target Datetime object
         */
        public HourState(final Duration date) {
            this.date = date;
        }

        /**
         * Parse the next token.
         *
         * @param c current token
         * @return next state
         */
        public IState handle(final String c) {
            if (c.matches("[0-9]")) {
                val = val.concat(c);
                return this;
            } else if (c.matches(":")) {
                date.setHour(Integer.parseInt(val));
                return new MinuteState(date);
            } else {
                throw new IllegalArgumentException();
            }
        }
    }

    /**
     * Minute parsing state.
     *
     * @author YDS
     *
     */
    private class MinuteState implements IState {
        /** date. */
        private final Duration date;
        /** string representation of the minutes. */
        private String val = "";

        /**
         * Constructor.
         *
         * @param date target Datetime object
         */
        public MinuteState(final Duration date) {
            this.date = date;
        }

        /**
         * Parse the next token.
         *
         * @param c current token
         * @return next state
         */
        public IState handle(final String c) {
            if (c.matches("[0-9]")) {
                val = val.concat(c);
                return this;
            } else if (c.matches("\\$")) {
                date.setMinute(Integer.parseInt(val));
                return new EndState();
            } else if (c.matches(":")) {
                date.setMinute(Integer.parseInt(val));
                return new SecondState(date);
            } else {
                throw new IllegalArgumentException();
            }
        }
    }

    /**
     * Second parsing state.
     *
     * @author YDS
     *
     */
    private class SecondState implements IState {
        /** date. */
        private final Duration date;
        /** string representation of the seconds. */
        private String val = "";

        /**
         * Constructor.
         *
         * @param date target Datetime object
         */
        public SecondState(final Duration date) {
            this.date = date;
        }

        /**
         * Parse the next token.
         *
         * @param c current token
         * @return next state
         */
        public IState handle(final String c) {
            if (c.matches("[0-9]")) {
                val = val.concat(c);
                return this;
            } else if (c.matches("\\$")) {
                date.setSecond(Integer.parseInt(val));
                return new EndState();
            } else {
                throw new IllegalArgumentException();
            }
        }
    }

    /**
     * Initial state.
     *
     * @author YDS
     *
     */
    private class EndState implements IState {

        /**
         * Parse the next token.
         *
         * @param c current token
         * @return next state
         */
        public IState handle(final String c) {
            throw new IllegalArgumentException();
        }
    }
}
