// Copyright 2013 SNCF. All rights reserved.
package com.sncf.checker.io;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.zip.ZipInputStream;

import com.sncf.checker.bo.Platform;
import com.sncf.checker.common.TreeMultiMap;

/**
 * The data source is a tabulation separated file containing the platforms.
 *
 * @author YDS
 *
 */
public class PlatformFlatFileDAO extends FlatFileDAO implements PlatformDAO {

    /** scenario file path. */
    private final String scenario;
    /** indexed by name platform map of the scenario. */
    private Map<String, Platform> cache = null;
    /** indexed by arrival/departure name platform multimap of the scenario. */
    private TreeMultiMap<String, Platform> dutyIndexedCompatible = null;

    /**
     * Constructor.
     *
     * @param scenario file path to the scenario zip archive
     * @param solution file path to the solution file
     */
    public PlatformFlatFileDAO(final String scenario, final String solution) {
        this.scenario = scenario;
    }

    /** TrainCategory DAO. */
    private TrainCategoryDAO trainCategoryDAO;

    /**
     * TrainCategory DAO setter.
     *
     * @param dao TrainCategory DAO
     */
    public void setTrainCategoryDAO(TrainCategoryDAO dao) {
        this.trainCategoryDAO = dao;
    }

    /** Fill DAO. */
    private FillDAO fillDAO;

    /**
     * Fill DAO setter.
     *
     * @param dao Fill DAO
     */
    public void setFillDAO(final FillDAO dao) {
        this.fillDAO = dao;
    }

    /** Gate DAO. */
    private GateDAO gateDAO;

    /**
     * Gate DAO setter.
     *
     * @param dao Gate DAO
     */
    public void setGateDAO(final GateDAO dao) {
        this.gateDAO = dao;
    }

    /**
     * Load a platform by name.
     *
     * @param name the searched platform name
     * @return platform, null if not found
     */
    public Platform loadByName(final String name) {
        initialize();
        Platform entity = cache.get(name);
        return entity;
    }

    /**
     * Load all platforms.
     *
     * @return platform list
     */
    public List<Platform> loadAll() {
        initialize();
        ArrayList<Platform> list = new ArrayList<Platform>();
        list.addAll(cache.values());
        return list;
    }

    /**
     * Extract all platforms from platforms.csv in zip file. Extract compatible platforms from prefPlat.csv.
     */
    private void initialize() {
        if ((cache == null) || (dutyIndexedCompatible == null)) {
            cache = new TreeMap<String, Platform>();

            BufferedReader bufferedReader = null;
            ZipInputStream zis = null;

            try {
                String current;

                String csvName = "platforms.csv";
                zis = getZipEntryInputStream(scenario, csvName);
                bufferedReader = new BufferedReader(new InputStreamReader(zis));

                bufferedReader.readLine();
                while ((current = bufferedReader.readLine()) != null) {
                    String[] splits = current.split(";");
                    Platform temp;

                    temp = deserialize(splits);
                    cache.put(temp.getName(), temp);
                }
            } catch (FileNotFoundException e) {
                e.printStackTrace();
            } catch (IOException e) {
                e.printStackTrace();
            } finally {
                if (bufferedReader != null) {
                    try {
                        bufferedReader.close();
                        zis.close();
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                }
            }

            dutyIndexedCompatible = new TreeMultiMap<String, Platform>();

            try {
                String current;

                String csvName = "prefPlat.csv";
                zis = getZipEntryInputStream(scenario, csvName);
                bufferedReader = new BufferedReader(new InputStreamReader(zis));

                bufferedReader.readLine();
                while ((current = bufferedReader.readLine()) != null) {
                    Platform platform;
                    String[] splits = current.split(";");

                    platform = loadByName(splits[1]);
                    dutyIndexedCompatible.put(splits[0], platform);
                }
            } catch (FileNotFoundException e) {
                e.printStackTrace();
            } catch (IOException e) {
                e.printStackTrace();
            } finally {
                if (bufferedReader != null) {
                    try {
                        bufferedReader.close();
                        zis.close();
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                }
            }
        }
    }

    /**
     * Load all platforms compatible with a given duty.
     *
     * @param name the duty name
     * @return platform list
     */
    public List<Platform> loadCompatible(final String name) {
        initialize();
        return dutyIndexedCompatible.get(name);
    }

    /**
     * Deserialize an entry in a Platform object.
     *
     * @param entry data extracted from the file
     * @return platform
     */
    private Platform deserialize(final String[] raw) {
        Platform entity = new Platform(gateDAO, fillDAO, trainCategoryDAO);

        entity.setName(raw[0]);
        entity.setLength(Integer.parseInt(raw[1]));

        return entity;
    }

}
