// Copyright 2013 SNCF. All rights reserved.
package com.sncf.checker.io;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.zip.ZipInputStream;

import com.sncf.checker.bo.Train;
import com.sncf.checker.common.Duration;

/**
 * The data source is a tabulation separated file containing the trains.
 *
 * @author YDS
 *
 */
public class TrainFlatFileDAO extends FlatFileDAO implements TrainDAO {

    /** scenario file path. */
    private final String scenario;
    /** indexed by name train map of the scenario. */
    private List<Train> orderedCache = null;
    /** indexed by name train map of the scenario. */
    private Map<String, Train> cache = null;

    /**
     * Constructor.
     *
     * @param scenario file path to the scenario zip archive
     * @param solution file path to the solution file
     */
    public TrainFlatFileDAO(final String scenario, final String solution) {
        this.scenario = scenario;
    }

    /** TrainCategory DAO. */
    private TrainCategoryDAO trainCategoryDAO;

    /**
     * TrainCategory DAO setter.
     *
     * @param dao TrainCategory DAO
     */
    public void setTrainCategoryDAO(final TrainCategoryDAO dao) {
        this.trainCategoryDAO = dao;
    }

    /** Fill DAO. */
    private FillDAO fillDAO;

    /**
     * Fill DAO setter.
     *
     * @param dao Fill DAO
     */
    public void setFillDAO(final FillDAO dao) {
        this.fillDAO = dao;
    }

    /** Arrival DAO. */
    private ArrivalDAO arrivalDAO;

    /**
     * Arrival DAO setter.
     *
     * @param dao Arrival DAO
     */
    public void setArrivalDAO(final ArrivalDAO dao) {
        this.arrivalDAO = dao;
    }

    /** Departure DAO. */
    private DepartureDAO departureDAO;

    /**
     * Departure DAO setter.
     *
     * @param dao Departure DAO
     */
    public void setDepartureDAO(final DepartureDAO dao) {
        this.departureDAO = dao;
    }

    /** Resource DAO. */
    private ResourceDAO resourceDAO;

    /**
     * Resource DAO setter.
     *
     * @param dao Resource DAO
     */
    public void setResourceDAO(final ResourceDAO dao) {
        this.resourceDAO = dao;
    }

    /**
     * Load a train by name.
     *
     * @param name the searched train name
     * @return train object, null if not found
     */
    public Train loadByName(final String name) {
        initialize();
        Train entity = cache.get(name);
        return entity;
    }

    /**
     * Load all trains.
     *
     * @return train list
     */
    public List<Train> loadAll() {
        initialize();
        ArrayList<Train> list = new ArrayList<Train>();
        list.addAll(orderedCache);
        return list;
    }

    /**
     * Extract all trains from initialTrains.csv in zip file.
     */
    private void initialize() {
        if (cache == null || orderedCache == null) {
            cache = new TreeMap<String, Train>();
            orderedCache = new ArrayList<Train>();

            BufferedReader bufferedReader = null;
            ZipInputStream zis = null;

            try {
                String current;

                String csvName = "arrivals.csv";
                zis = getZipEntryInputStream(scenario, csvName);

                bufferedReader = new BufferedReader(new InputStreamReader(zis));
                bufferedReader.readLine();
                while ((current = bufferedReader.readLine()) != null) {
                    String[] splits = current.split(";", -1);
                    Train temp;

                    temp = deserializeFromArrivals(splits);
                    cache.put(temp.getName(), temp);
                    orderedCache.add(temp);
                    temp.setArrival(arrivalDAO.loadByTrain(temp.getName()));
                    temp.setCategory(trainCategoryDAO.loadByName(splits[6]));
                    temp.setDeparture(departureDAO.loadByTrain(temp.getName()));
                }
            } catch (FileNotFoundException e) {
                e.printStackTrace();
            } catch (IOException e) {
                e.printStackTrace();
            } finally {
                if (bufferedReader != null) {
                    try {
                        bufferedReader.close();
                        zis.close();
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                }
            }

            try {
                String current;

                String csvName = "initialTrains.csv";
                zis = getZipEntryInputStream(scenario, csvName);

                bufferedReader = new BufferedReader(new InputStreamReader(zis));
                bufferedReader.readLine();
                while ((current = bufferedReader.readLine()) != null) {
                    String[] splits = current.split(";", -1);
                    Train temp;

                    temp = deserializeFromInitialTrains(splits);
                    cache.put(temp.getName(), temp);
                    orderedCache.add(temp);
                    temp.setCategory(trainCategoryDAO.loadByName(splits[1]));
                    temp.setResource(resourceDAO.loadByName(splits[2]));
                    temp.setDeparture(departureDAO.loadByTrain(temp.getName()));
                }
            } catch (FileNotFoundException e) {
                e.printStackTrace();
            } catch (IOException e) {
                e.printStackTrace();
            } finally {
                if (bufferedReader != null) {
                    try {
                        bufferedReader.close();
                        zis.close();
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                }
            }
        }
    }

    /**
     * Deserialize an entry from InitialTrain.txt in a Train object.
     *
     * @param entry data extracted from the file
     * @return train
     */
    private Train deserializeFromInitialTrains(final String[] raw) {
        Train entity = new Train(fillDAO);

        entity.setName(raw[0]);
        entity.setRemDBM(Integer.parseInt(raw[3]));
        entity.setRemTBM(new Duration(raw[4]));

        return entity;
    }

    /**
     * Deserialize an entry from Arrivals.txt in a Train object.
     *
     * @param entry data extracted from the file
     * @return train
     */
    private Train deserializeFromArrivals(final String[] raw) {
        Train entity = new Train(fillDAO);

        entity.setName(raw[1]);
        entity.setRemDBM(Integer.parseInt(raw[9]));
        entity.setRemTBM(new Duration(raw[10]));

        return entity;
    }
}
